# =============================================================
#
# Demo code for training Wasserstein RBMs based on the paper
# 
#  Wasserstein Training of Restricted Boltzmann Machines
#  G. Montavon, KR Mueller, M. Cuturi, NIPS 2016
#
#  Example of execution:
#  > python train.py
#
# =============================================================

import numpy,scipy,scipy.spatial

na = numpy.newaxis

def arcsigm(x):   return numpy.arctanh(2*x-1)*2
def sigm(x):      return numpy.tanh(0.5*x)*0.5+0.5
def softplus(x):  return numpy.log(1+numpy.exp(numpy.minimum(x,5)))+numpy.maximum(x-5,0)
def realize(x,rstate): return (x > rstate.uniform(0,1,x.shape))*1.0

# -----------------------------------------
# Smooth Wasserstein distance computation
# -----------------------------------------

metric = 'sqeuclidean' # (same as Hamming for binary data)
#stopcriterion = 'wsddecrease' # used in the paper
stopcriterion = 'dualvariabledecrease'


def T(K,u,v): return u[:,na]*v[na,:]*K

def wasserstein(XA,XB,L):

	D = scipy.spatial.distance.cdist(XA,XB,metric)

	K = numpy.exp(-L*D)
	
	KD = K*D

	np,mb = K.shape

	u = numpy.ones([np]) / np
	v=1./(mb*numpy.dot(K.T,u))
	
	oldWSD = numpy.dot(u,numpy.dot(KD,v))

	for i in range(1,10000):
		u=1./(np*numpy.dot(K  ,v))
		v=1./(mb*numpy.dot(K.T,u))
		if stopcriterion == 'wsddecrease':
			WSD = numpy.dot(u,numpy.dot(KD,v))
			if WSD - oldWSD < 0.001: break
			else: oldWSD = WSD
		elif stopcriterion == 'dualvariabledecrease' and i % 20 == 0:
			marginold = u*numpy.dot(K,v)
			u=1./(np*numpy.dot(K  ,v))
			v=1./(mb*numpy.dot(K.T,u))
			marginnew = u*numpy.dot(K,v)
			if numpy.linalg.norm((marginold -marginnew), ord=1)< 1e-3:
				WSD = numpy.dot(u,numpy.dot(KD,v))
				break

	alpha = numpy.log(u)/L
	beta  = numpy.log(v)/L

	return alpha-alpha.mean(),beta-beta.mean(),WSD

# -----------------------------------------
# Restricted Boltzmann Machine
# -----------------------------------------

class RBM:

	def __init__(self,X,n,L):

		# base parameters
		self.A0 = arcsigm(numpy.clip(X.mean(axis=0),0.01,0.99))
		self.B0 = numpy.zeros([n])

		# parameters to be learned
		self.W = numpy.zeros([X.shape[1],n])
		self.A = self.A0*1
		self.B = self.B0*1

		# parameter gradients
		self.DW = self.W*0
		self.DA = self.A*0
		self.DB = self.B*0

		# calibrate WSD scale to the data
		self.L = L/scipy.spatial.distance.cdist(X[:100],X[:100],'sqeuclidean').mean()

		# initialize state
		self.Xd = X*1
		self.Xm = X*0

	def sample(self,rstate):

		# sample data-dependent hidden states
		self.Yd = realize(sigm(numpy.dot(self.Xd,self.W  )+self.B),rstate)

		# perform one step of PCD
		self.Ym = realize(sigm(numpy.dot(self.Xm,self.W  )+self.B),rstate)
		self.Xm = realize(sigm(numpy.dot(self.Ym,self.W.T)+self.A),rstate)

	def gradWSD(self):

		np = len(self.Xm)

		# compute the Wasserstein gradient and normalize it
		alpha,beta,_ = wasserstein(self.Xm,self.Xd,self.L)
		alpha = alpha / alpha.std() * 2**.5

		# propagate Wasserstein gradient to RBM
		a = self.Xd.mean(axis=0)
		b = self.Yd.mean(axis=0)
		DWe = (numpy.dot(alpha*(self.Xm-a).T,self.Ym-b))/np
		DAe = ((alpha[:,na]*(self.Xm-a)).sum(axis=0))/np
		DBe = ((alpha[:,na]*(self.Ym-b)).sum(axis=0))/np

		# gradient centering
		self.DW += DWe
		self.DA += (DAe - numpy.dot(DWe,b))
		self.DB += (DBe - numpy.dot(a,DWe))

	def gradKL(self):

		mb = len(self.Xd)
		np = len(self.Xm)

		# compute KL gradient
		a = self.Xd.mean(axis=0)
		b = self.Yd.mean(axis=0)
		DWe = numpy.dot((self.Xm-a).T,(self.Ym-b))/np - numpy.dot((self.Xd-a).T,(self.Yd-b))/mb
		DAe = (self.Xm-a).sum(axis=0)/np - (self.Xd-a).sum(axis=0)/mb
		DBe = (self.Ym-b).sum(axis=0)/np - (self.Yd-b).sum(axis=0)/mb

		# gradient centering
		self.DW += self.klreg * DWe
		self.DA += self.klreg * (DAe - numpy.dot(DWe,b))
		self.DB += self.klreg * (DBe - numpy.dot(a,DWe))

	def gradFROB(self):

		self.DW += self.frobreg * self.W
		self.DA += self.frobreg * (self.A-self.A0)

	def update(self):

		self.W = self.W - self.lr*self.DW; self.DW*=0
		self.A = self.A - self.lr*self.DA; self.DA*=0
		self.B = self.B - self.lr*self.DB; self.DB*=0

	def FE(self,X,t):

		# compute RBM free energy
		W,A,B = self.W*t,self.A*t+self.A0*(1-t),self.B*t+self.B0*(1-t)
		return  - numpy.dot(X,A) - softplus(numpy.dot(X,W)+B).sum(axis=1)

	def AIS(self,rstate,T,n):

		# apply AIS procedure
		XGEN,LOGWAIS = numpy.zeros([n,len(self.A)]),0
		for t1,t2 in zip(T[:-1],T[1:]):
			W,A,B = self.W*t1,self.A*t1+self.A0*(1-t1),self.B*t1+self.B0*(1-t1)
			XGEN = realize(sigm(numpy.dot(realize(sigm(numpy.dot(XGEN,W)+B),rstate),W.T)+A),rstate)
			LOGWAIS += self.FE(XGEN,t1)-self.FE(XGEN,t2)

		return XGEN,LOGWAIS

	def NLL(self,X,LOGWAIS):

		# compute the data negative log-likelihood (based on AIS result)
		LOGZ1Z0 = numpy.log(numpy.exp(LOGWAIS-LOGWAIS.max()).mean())+LOGWAIS.max()
		LOGZ0 = softplus(self.A0).sum() + softplus(self.B0).sum()
		NLL = -(-self.FE(X,1)-LOGZ0-LOGZ1Z0).mean()
		STDZ = LOGWAIS.std()
		return NLL,STDZ

	def WSD(self,XGEN,X):

		alpha,beta,WSD = wasserstein(XGEN,X,self.L)
		return WSD

