# =============================================================
#
# Demo code for training Wasserstein RBMs based on the paper
# 
#  Wasserstein Training of Restricted Boltzmann Machines
#  G. Montavon, KR Mueller, M. Cuturi, NIPS 2016
#
#  Example of execution:
#  > python train.py
#
# =============================================================

import numpy,modules,data,scipy,scipy.io

# -----------------------------------------
# Hyperparameters
# -----------------------------------------

L       = 10     # Wasserstein scale parameter
lrpre   = 0.250  # base learning rate
klreg   = 1      # KL regularization (0-4)
frobreg = 0      # Frobenius regularization (0-2)

# -----------------------------------------
# Create RBM
# -----------------------------------------
rbm = modules.RBM(data.Xtrain,400,L)
rbm.klreg   = [1e-9,0.1,1.0,10.0,1e9][klreg]
rbm.frobreg = [0.0001,0.001,0.01][frobreg]*rbm.klreg

# -----------------------------------------
# Training procedure
# -----------------------------------------

print("[  'it' ,    'W' ,    'A' ,    'B' ,   'NLLTR' ,   'NLLVA' ,   'NLLTT' ,   'WSDTR' ,   'WSDVA' ,   'WSDTT' ],")

# apply 2500 iterations of pretraining and 10000 iterations of training
for it in range(-2500,10001):
	
	# adapt learning rate to pretraining/training phase
	if it < 0: rbm.lr =            lrpre/rbm.klreg
	else:      rbm.lr = min(0.0025,0.025/rbm.klreg)

	# perform one step of gradient descent
	rbm.sample(numpy.random.mtrand.RandomState(it+2500))
	rbm.gradKL()
	rbm.gradFROB()
	if it >= 0: rbm.gradWSD()
	rbm.update()

	# verify what has been learned every 250 iterations
	if it%250 == 0:

		# estimate log-likelihood (using AIS)
		_,LOGWAIS = rbm.AIS(numpy.random.mtrand.RandomState(0),1-(1-numpy.arange(0,1+1e-9,0.001))**2,100.0)
		NLLtr,_ = rbm.NLL(data.Xtrain,LOGWAIS)
		NLLva,_ = rbm.NLL(data.Xvalid,LOGWAIS)
		NLLtt,_ = rbm.NLL(data.Xtest, LOGWAIS)

		# compute the smooth Wasserstein distance (approximated from PCD chains)
		WSDtr = rbm.WSD(rbm.Xm,data.Xtrain)
		WSDva = rbm.WSD(rbm.Xm,data.Xvalid)
		WSDtt = rbm.WSD(rbm.Xm,data.Xtest)

		# display current status of training
		print('[ %5d , %6.3f , %6.3f , %6.3f , %9.3f , %9.3f , %9.3f , %9.3f , %9.3f , %9.3f ],'%(
			it,rbm.W.std(),rbm.A.mean(),rbm.B.mean(),
			numpy.clip(NLLtr,-9999,9999),numpy.clip(NLLva,-9999,9999),numpy.clip(NLLtt,-9999,9999),
			numpy.clip(WSDtr,-9999,9999),numpy.clip(WSDva,-9999,9999),numpy.clip(WSDtt,-9999,9999)))

		scipy.io.savemat('rbm.mat',{'W':rbm.W,'B':rbm.B,'A':rbm.A,'X':rbm.Xm})

